@@ -27,8 +27,8 @@ from .__version__ import (
     ascmhl_default_hashformat,
 )
 from .generator import MHLGenerationCreationSession
-from .hasher import create_filehash, DirectoryHashContext
-from .hashlist import MHLCreatorInfo, MHLProcessInfo, MHLTool, MHLProcess
+from .hasher import hash_file, DirectoryHashContext
+from .hashlist import MHLMediaHash, MHLCreatorInfo, MHLProcessInfo, MHLTool, MHLProcess
 from .history import MHLHistory
 from .traverse import post_order_lexicographic
 
@@ -116,7 +116,7 @@ def create_for_folder_subcommand(
     if not os.path.isabs(root_path):
         root_path = os.path.join(os.getcwd(), root_path)
 
-    logger.verbose(f"Sealing folder at path: {root_path} ...")
+    logger.verbose(f"Creating new generation for folder at path: {root_path} ...")
 
     existing_history = MHLHistory.load_from_path(root_path)
 
@@ -261,10 +261,49 @@ def create_for_single_files_subcommand(
     "-dh",
     default=False,
     is_flag=True,
-    help="Record single file, no completeness check (multiple occurrences possible for adding multiple files",
+    help="Just create and verify directory hashes.",
 )
-@click.option("--hash_format", "-h", type=click.Choice(ascmhl_supported_hashformats), multiple=False, help="Algorithm")
-def verify(root_path, verbose, directory_hash, hash_format, ignore_list, ignore_spec_file):
+# options for directory_hash
+@click.option(
+    "--calculate_only",
+    "-co",
+    default=False,
+    is_flag=True,
+    help="Only calculate and print hashes, don't compare against history.",
+)
+@click.option(
+    "--root_only",
+    "-ro",
+    default=False,
+    is_flag=True,
+    help=(
+        "Only calculate and print root directory hash, don't compare against history (only applies to"
+        " --directory_hash)."
+    ),
+)
+# options
+@click.option(
+    "--hash_format",
+    "-h",
+    type=click.Choice(ascmhl_supported_hashformats),
+    multiple=False,
+    help="Algorithm for directory hashes",
+)
+# subcommand
+@click.option(
+    "--packing_list", "-pl", default=None, type=click.Path(exists=True), help="Verify against an external packing list"
+)
+def verify(
+    root_path,
+    verbose,
+    directory_hash,
+    hash_format,
+    packing_list,
+    ignore_list,
+    ignore_spec_file,
+    calculate_only,
+    root_only,
+):
     """
     Verify a folder, single file(s), or a directory hash
 
@@ -275,25 +314,34 @@ def verify(root_path, verbose, directory_hash, hash_format, ignore_list, ignore_
     files in the file system are reported as errors. No new ASC MHL file /
     generation is created.
     """
-    if directory_hash is True:
-        verify_directory_hash_subcommand(root_path, verbose, hash_format, ignore_list, ignore_spec_file)
+
+    if packing_list is not None:
+        verify_entire_folder(root_path, verbose, packing_list, ignore_list, ignore_spec_file, calculate_only)
         return
 
-    verify_entire_folder_against_full_history_subcommand(root_path, verbose, ignore_list, ignore_spec_file)
+    if directory_hash is True:
+        verify_directory_hash_subcommand(
+            root_path, verbose, hash_format, ignore_list, ignore_spec_file, calculate_only, root_only
+        )
+        return
+
+    verify_entire_folder(root_path, verbose, None, ignore_list, ignore_spec_file)
     return
 
 
-def verify_entire_folder_against_full_history_subcommand(root_path, verbose, ignore_list=None, ignore_spec_file=None):
-    # command formerly known as "check"
+def verify_entire_folder(
+    root_path, verbose, packing_list_path, ignore_list=None, ignore_spec_file=None, calculate_only=None
+):
     """
-    Checks MHL hashes from all generations against all file hashes.
+    Checks MHL hashes from all generations / a packing list against all file hashes.
 
     ROOT_PATH: the root path to use for the asc mhl history
+    packing_list (opt): the path to the packing list
 
     Traverses through the content of a folder, hashes all found files and compares ("verifies") the hashes
-    against the records in the asc-mhl folder. The command finds all files that are existent in the file system
-    but are not registered in the `asc-mhl` folder yet, and all files that are registered in the `asc-mhl` folder
-    but that are missing in the file system.
+    against the records in the asc-mhl folder/packing list. The command finds all files that are existent in
+    the file system but are not registered in the `asc-mhl` folder/packing list yet, and all files that are
+    registered in the `asc-mhl` folder/packing list but that are missing in the file system.
     """
     logger.verbose_logging = verbose
 
@@ -302,7 +350,10 @@ def verify_entire_folder_against_full_history_subcommand(root_path, verbose, ign
 
     logger.verbose(f"check folder at path: {root_path}")
 
-    existing_history = MHLHistory.load_from_path(root_path)
+    if packing_list_path is not None:
+        existing_history = MHLHistory.load_from_packing_list_path(packing_list_path, root_path)
+    else:
+        existing_history = MHLHistory.load_from_path(root_path)
 
     if len(existing_history.hash_lists) == 0:
         raise errors.NoMHLHistoryException(root_path)
@@ -336,9 +387,9 @@ def verify_entire_folder_against_full_history_subcommand(root_path, verbose, ign
                 continue
 
             # create a new hash and compare it against the original hash entry
-            current_hash = create_filehash(original_hash_entry.hash_format, file_path)
+            current_hash = hash_file(file_path, original_hash_entry.hash_format)
             if original_hash_entry.hash_string == current_hash:
-                logger.verbose(f"verification of file {relative_path}: OK")
+                logger.verbose(f"verification ({original_hash_entry.hash_format}) of file {relative_path}: OK")
             else:
                 logger.error(
                     f"ERROR: hash mismatch        for {relative_path} "
@@ -357,7 +408,9 @@ def verify_entire_folder_against_full_history_subcommand(root_path, verbose, ign
         raise exception
 
 
-def verify_directory_hash_subcommand(root_path, verbose, hash_format, ignore_list=None, ignore_spec_file=None):
+def verify_directory_hash_subcommand(
+    root_path, verbose, hash_format, ignore_list=None, ignore_spec_file=None, calculate_only=False, root_only=False
+):
     """
     Checks MHL directory hashes from all generations against computed directory hashes.
 
@@ -424,21 +477,27 @@ def verify_directory_hash_subcommand(root_path, verbose, hash_format, ignore_lis
                     if directory_hash_entry.hash_format != hash_format:
                         continue
                     found_hash_format = True
-                    num_current_successful_verifications = _compare_and_log_directory_hashes(
-                        relative_path, directory_hash_entry, content_hash, structure_hash
-                    )
+
+                    num_current_successful_verifications = -1
+                    if not root_only:
+                        num_current_successful_verifications = _compare_and_log_directory_hashes(
+                            relative_path, directory_hash_entry, content_hash, structure_hash
+                        )
+
                     if num_current_successful_verifications == 2:
                         num_successful_verifications += 1
                     if num_current_successful_verifications == 1:
                         num_failed_verifications += 1
 
-                if not found_hash_format:
-                    logger.error(
-                        f"ERROR: verification of folder {relative_path}: No directory hash of type {hash_format} found"
-                    )
-                    num_failed_verifications += 1
+                if not calculate_only:
+                    if not found_hash_format:
+                        logger.error(
+                            f"ERROR: verification of folder {relative_path}: No directory hash of type"
+                            f" {hash_format} found"
+                        )
+                        num_failed_verifications += 1
             else:
-                hash_string = hash_file_path(existing_history, file_path, hash_format, session)
+                hash_string = hash_file(file_path, hash_format)
                 dir_hash_context.append_file_hash(file_path, hash_string)
         dir_content_hash = None
         dir_structure_hash = None
@@ -448,9 +507,15 @@ def verify_directory_hash_subcommand(root_path, verbose, hash_format, ignore_lis
             dir_content_hash_mappings[folder_path] = dir_content_hash
             dir_structure_hash_mappings[folder_path] = dir_structure_hash
         modification_date = datetime.datetime.fromtimestamp(os.path.getmtime(folder_path))
+
+        logger.verbose_logging = calculate_only
+        relative_path = session.root_history.get_relative_file_path(folder_path)
+        if root_only and relative_path != ".":
+            logger.verbose_logging = verbose
         session.append_directory_hashes(
             folder_path, modification_date, hash_format, dir_content_hash, dir_structure_hash
         )
+        logger.verbose_logging = verbose
 
         # compare root hashes, works differently
         if folder_path == root_path:
@@ -464,8 +529,9 @@ def verify_directory_hash_subcommand(root_path, verbose, hash_format, ignore_lis
                                 ".", root_hash_entry, dir_content_hash, dir_structure_hash
                             )
                             found_hash_format = True
-            if not found_hash_format:
-                logger.error(f"ERROR: verification of root folder: No directory hash of type {hash_format} found")
+            if not calculate_only:
+                if not found_hash_format:
+                    logger.error(f"ERROR: verification of root folder: No directory hash of type {hash_format} found")
 
     exception = None
     if num_failed_verifications > 0:
@@ -631,6 +697,101 @@ def diff_entire_folder_against_full_history_subcommand(root_path, verbose, ignor
         raise exception
 
 
+@click.command()
+@click.argument("root_path", type=click.Path(exists=True))
+@click.argument("destination_path", type=click.Path())
+# general options
+@click.option("--verbose", "-v", default=False, is_flag=True, help="Verbose output")
+@click.option(
+    "--no_directory_hashes",
+    "-n",
+    default=False,
+    is_flag=True,
+    help="Skip creation of directory hashes, only reference directories without hash",
+)
+@click.option("ignore_list", "--ignore", "-i", multiple=True, help="A single file pattern to ignore.")
+@click.option(
+    "ignore_spec_file",
+    "--ignore_spec",
+    "-ii",
+    type=click.Path(exists=True),
+    help="A file containing multiple file patterns to ignore.",
+)
+def flatten(root_path, destination_path, verbose, no_directory_hashes, ignore_list, ignore_spec_file):
+    """
+    Flatten an MHL history into one external manifest
+
+    \b
+    The flatten command iterates through the mhl-history, collects all known files and
+    their hashes in multiple hash formats and writes them to a new mhl file outside of the
+    iterated history.
+    """
+    flatten_history(root_path, destination_path, verbose, no_directory_hashes, ignore_list, ignore_spec_file)
+    return
+
+
+def flatten_history(root_path, destination_path, verbose, no_directory_hashes, ignore_list=None, ignore_spec_file=None):
+    logger.verbose_logging = verbose
+
+    if not os.path.isabs(root_path):
+        root_path = os.path.join(os.getcwd(), root_path)
+
+    logger.verbose(f"Flattening folder at path: {root_path} ...")
+
+    existing_history = MHLHistory.load_from_path(root_path)
+
+    # create the ignore specification
+    ignore_spec = ignore.MHLIgnoreSpec(existing_history.latest_ignore_patterns(), ignore_list, ignore_spec_file)
+
+    # start a verification session on the existing history
+    collection_history = MHLHistory.create_collection_at_path(destination_path)
+    session = MHLGenerationCreationSession(collection_history, ignore_spec)
+
+    # store the directory hashes of sub folders so we can use it when calculating the hash of the parent folder
+    dir_hash_mappings = {}
+
+    if len(existing_history.hash_lists) == 0:
+        raise errors.NoMHLHistoryException(root_path)
+
+    for hash_list in existing_history.hash_lists:
+        for media_hash in hash_list.media_hashes:
+            if not media_hash.is_directory:
+                for hash_entry in media_hash.hash_entries:
+                    if hash_entry.action != "failed":
+                        # check if this entry is newer than the one already in there to avoid duplicate entries
+                        found_media_hash = session.new_hash_lists[collection_history].find_media_hash_for_path(
+                            media_hash.path
+                        )
+                        if found_media_hash == None:
+                            session.append_file_hash(
+                                media_hash.path,
+                                media_hash.file_size,
+                                media_hash.last_modification_date,
+                                hash_entry.hash_format,
+                                hash_entry.hash_string,
+                                action=hash_entry.action,
+                                hash_date=hash_entry.hash_date,
+                            )
+                        else:
+                            hashformat_is_already_there = False
+                            for found_hash_entry in found_media_hash.hash_entries:
+                                if found_hash_entry.hash_format == hash_entry.hash_format:
+                                    hashformat_is_already_there = True
+                            if not hashformat_is_already_there:
+                                # assuming that hash_entry of same type also has same hash_value ..
+                                session.append_file_hash(
+                                    media_hash.path,
+                                    media_hash.file_size,
+                                    media_hash.last_modification_date,
+                                    hash_entry.hash_format,
+                                    hash_entry.hash_string,
+                                    action=hash_entry.action,
+                                    hash_date=hash_entry.hash_date,
+                                )
+
+    commit_session_for_collection(session, root_path)
+
+
 @click.command()
 @click.option(
     "--verbose",
@@ -723,7 +884,15 @@ def info_for_single_file(root_path, verbose, single_file):
 
 @click.command()
 @click.argument("file_path", type=click.Path(exists=True))
-def xsd_schema_check(file_path):
+# subcommands
+@click.option(
+    "--directory_file",
+    "-df",
+    default=False,
+    is_flag=True,
+    help="Check directory file (e.g. ascmhl_chain.xml) instead of manifest file",
+)
+def xsd_schema_check(file_path, directory_file):
     """
     Checks a .mhl file against the xsd schema definition
 
@@ -735,6 +904,10 @@ def xsd_schema_check(file_path):
     """
 
     xsd_path = "xsd/ASCMHL.xsd"
+
+    if directory_file:
+        xsd_path = "xsd/ASCMHLDirectory__combined.xsd"
+
     xsd = etree.XMLSchema(etree.parse(xsd_path))
 
     # pass a file handle to support the fake file system used in the tests
@@ -749,69 +922,6 @@ def xsd_schema_check(file_path):
         raise errors.VerificationFailedException
 
 
-# TODO should be part of the `verify -dh` subcommand
-@click.command()
-@click.argument("root_path", type=click.Path(exists=True))
-@click.option(
-    "--verbose",
-    "-v",
-    default=False,
-    is_flag=True,
-    help="Print all directory hashes of sub directories",
-)
-@click.option(
-    "--hash_format",
-    "-h",
-    type=click.Choice(ascmhl_supported_hashformats),
-    multiple=False,
-    default=ascmhl_default_hashformat,
-    help="Algorithm",
-)
-@click.option(
-    "ignore_list",
-    "--ignore",
-    "-i",
-    multiple=True,
-    help="A single file pattern to ignore.",
-)
-@click.option(
-    "ignore_spec_file",
-    "--ignore_spec",
-    "-ii",
-    type=click.Path(exists=True),
-    help="A file containing multiple file patterns to ignore.",
-)
-def directory_hash(root_path, verbose, hash_format, ignore_list, ignore_spec_file):
-    """
-    [TMP] Creates the directory hash of a given folder
-    """
-    if not os.path.isabs(root_path):
-        root_path = os.path.join(os.getcwd(), root_path)
-
-    # store the directory hashes of sub folders so we can use it when calculating the hash of the parent folder
-    dir_hash_mappings = {}
-
-    ignore_spec = ignore.MHLIgnoreSpec(None, ignore_list, ignore_spec_file)
-
-    for folder_path, children in post_order_lexicographic(root_path, ignore_spec.get_path_spec()):
-        dir_hash_context = DirectoryHashContext(hash_format)
-        for item_name, is_dir in children:
-            item_path = os.path.join(folder_path, item_name)
-            if is_dir:
-                if not dir_hash_context:
-                    continue
-                hash_string = dir_hash_mappings.pop(item_path)
-            else:
-                hash_string = create_filehash(hash_format, item_path)
-            dir_hash_context.append_hash(hash_string, item_name)
-        dir_hash = dir_hash_context.final_hash_str()
-        dir_hash_mappings[folder_path] = dir_hash
-        if folder_path == root_path:
-            logger.info(f"  calculated root hash: {hash_format}: {dir_hash}")
-        elif verbose:
-            logger.info(f"directory hash for: {folder_path} {hash_format}: {dir_hash}")
-
-
 def test_for_missing_files(not_found_paths, root_path, ignore_spec: MHLIgnoreSpec = MHLIgnoreSpec()):
     ignore_path_spec = ignore_spec.get_path_spec()
     # update to exclude our ignored files
@@ -835,6 +945,20 @@ def commit_session(session):
     session.commit(creator_info, process_info)
 
 
+def commit_session_for_collection(session, root_path):
+    creator_info = MHLCreatorInfo()
+    creator_info.tool = MHLTool(ascmhl_tool_name, ascmhl_tool_version)
+    creator_info.creation_date = utils.datetime_now_isostring()
+    creator_info.host_name = platform.node()
+    process_info = MHLProcessInfo()
+    process_info.process = MHLProcess("flatten")
+    root_hash = MHLMediaHash()
+    root_hash.path = root_path
+    process_info.root_media_hash = root_hash
+    process_info.hashlist_custom_basename = "packinglist_" + os.path.basename(root_path)
+    session.commit(creator_info, process_info)
+
+
 def seal_file_path(existing_history, file_path, hash_format, session) -> (str, bool):
     relative_path = existing_history.get_relative_file_path(file_path)
     file_size = os.path.getsize(file_path)
@@ -850,24 +974,16 @@ def seal_file_path(existing_history, file_path, hash_format, session) -> (str, b
     success = True
     if len(existing_hash_formats) > 0 and hash_format not in existing_hash_formats:
         existing_hash_format = existing_hash_formats[0]
-        hash_in_existing_format = create_filehash(existing_hash_format, file_path)
+        hash_in_existing_format = hash_file(file_path, existing_hash_format)
         # FIXME: test what happens if the existing hash verification fails in other format fails
         # should we then really create two entries
         success &= session.append_file_hash(
             file_path, file_size, file_modification_date, existing_hash_format, hash_in_existing_format
         )
-    current_format_hash = create_filehash(hash_format, file_path)
+    current_format_hash = hash_file(file_path, hash_format)
     # in case the existing hash verification failed we don't want to add the current format hash to the generation
     # but we need to return it for directory hash creation
     if not success:
         return current_format_hash, False
     success &= session.append_file_hash(file_path, file_size, file_modification_date, hash_format, current_format_hash)
     return current_format_hash, success
-
-
-def hash_file_path(existing_history, file_path, hash_format, session) -> (str):
-    current_format_hash = create_filehash(hash_format, file_path)
-    relative_path = session.root_history.get_relative_file_path(file_path)
-    logger.verbose(f"  created file hash for         {relative_path}")
-
-    return current_format_hash
